# This package will contain the spiders of your Scrapy project
#
# Please refer to the documentation for information on how to create and manage
# your spiders.

import scrapy

from lich_linkextractor.items import ExampleItem


class ExampleSpider(scrapy.Spider):
    """ ExampleSpider
    Auto generated by os-scrapy-cookiecuter

    Run:
        scrapy crawl example
    """

    name = "example"

    def start_requests(self):
        yield scrapy.Request(
            url="https://www.example.com/",
            headers={"Referer":"http://www.123.com"},
            meta={
                # "depth": 0,
                "extractor.depth_limit": 1,
                # "extractor.redirect.switch": True,
                "extractor.rules": [
                    {
                        
                        "same_domain_only": False,
                        "allow": [],
                        "allow_domains": [],
                        "deny": [],
                        "deny_domains": [],
                        "restrict_xpaths": [],
                        "type": "lxml",
                    }
                ],
                # "extractor2.template": {
                #     "isRoot": True,
                #     "method": "table",
                #     "listData": [
                #         {
                #             "id": "92147549961982",
                #             "listName": "xingming",
                #             "method": "text",
                #             "rawSelector": ".neiye > .Survey > .Search_box > .Expert_xq > .Expert_xq11",
                #             "required": True,
                #             "selector": ".Expert_xq11",
                #             "xpath": './/*[contains(concat(" ",normalize-space(@class)," ")," Expert_xq11 ")]',
                #         },
                #         {
                #             "id": "92148240689730",
                #             "listName": "jianjie",
                #             "method": "text",
                #             "rawSelector": ".neiye > .Survey > .Search_box > .Expert_xq > ul > li > span > span",
                #             "required": True,
                #             "selector": "ul > li > span > span",
                #             "xpath": ".//ul/li/span/span",
                #         },
                #     ],
                #     "selector": ".neiye > .Survey > .Search_box > .Expert_xq",
                #     "xpath": './/*[contains(concat(" ",normalize-space(@class)," ")," neiye ")]/*[contains(concat(" ",normalize-space(@class)," ")," Survey ")]/*[contains(concat(" ",normalize-space(@class)," ")," Search_box ")]/*[contains(concat(" ",normalize-space(@class)," ")," Expert_xq ")]',
                # },
                "render_proxy.api": "http://render-proxy.thanos.sogou/RenderService/Render",
                "render_proxy.disableCacheserver": False,
                "render_proxy.downloadCss": False,
                "render_proxy.downloadPic": False,
                "render_proxy.enableJs": False,
                "render_proxy.loadIframe": False,
                "render_proxy.loadOtherResource": False,
                "render_proxy.needPosition": False,
                "render_proxy.needScreenshot": False,
            },
        )

    def parse(self, response):
        yield ExampleItem(
            url=response.url,
            request_headers=response.request.headers,
            response_headers=response.headers,
            status=response.status,
            meta=response.meta,
            body=response.body,
        )



# -*- coding: utf-8 -*-

# Scrapy settings for lich_linkextractor project
#
# For simplicity, this file contains only settings considered important or
# commonly used. You can find more settings consulting the documentation:
#
#     https://docs.scrapy.org/en/latest/topics/settings.html
#     https://docs.scrapy.org/en/latest/topics/downloader-middleware.html
#     https://docs.scrapy.org/en/latest/topics/spider-middleware.html

BOT_NAME = "lich_linkextractor"

SPIDER_MODULES = ["lich_linkextractor.spiders"]
NEWSPIDER_MODULE = "lich_linkextractor.spiders"


# Crawl responsibly by identifying yourself (and your website) on the user-agent
# USER_AGENT = 'lich_linkextractor (+http://www.yourdomain.com)'

# Obey robots.txt rules
ROBOTSTXT_OBEY = False
TWISTED_REACTOR = "twisted.internet.asyncioreactor.AsyncioSelectorReactor"
# Configure maximum concurrent requests performed by Scrapy (default: 16)
# CONCURRENT_REQUESTS = 32

# Configure a delay for requests for the same website (default: 0)
# See https://docs.scrapy.org/en/latest/topics/settings.html#download-delay
# See also autothrottle settings and docs
# DOWNLOAD_DELAY = 3
# The download delay setting will honor only one of:
# CONCURRENT_REQUESTS_PER_DOMAIN = 16
# CONCURRENT_REQUESTS_PER_IP = 16

# Disable cookies (enabled by default)
# COOKIES_ENABLED = False

# Disable Telnet Console (enabled by default)
# TELNETCONSOLE_ENABLED = False

# Override the default request headers:
# DEFAULT_REQUEST_HEADERS = {
#   'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
#   'Accept-Language': 'en',
# }

# Enable or disable spider middlewares
# See https://docs.scrapy.org/en/latest/topics/spider-middleware.html
# SPIDER_MIDDLEWARES = {
#    'lich_linkextractor.middlewares.LichLinkextractorSpiderMiddleware': 543,
# }

# Enable or disable downloader middlewares
# See https://docs.scrapy.org/en/latest/topics/downloader-middleware.html
DOWNLOADER_MIDDLEWARES = {
    "lich_linkextractor.middlewares.LinkExtractorDownloaderMiddleware": 545,
    "render_proxy_scrapy.middlewares.RenderProxyResponseDownloaderMiddleware": 555,
    "render_proxy_scrapy.middlewares.RenderProxyRequestDownloaderMiddleware": 800,
}

# Enable or disable extensions
# See https://docs.scrapy.org/en/latest/topics/extensions.html
# EXTENSIONS = {
#    'scrapy.extensions.telnet.TelnetConsole': None,
# }

# Configure item pipelines
# See https://docs.scrapy.org/en/latest/topics/item-pipeline.html
# ITEM_PIPELINES = {
#    'lich_linkextractor.pipelines.LichLinkextractorPipeline': 300,
# }

# Enable and configure the AutoThrottle extension (disabled by default)
# See https://docs.scrapy.org/en/latest/topics/autothrottle.html
# AUTOTHROTTLE_ENABLED = True
# The initial download delay
# AUTOTHROTTLE_START_DELAY = 5
# The maximum download delay to be set in case of high latencies
# AUTOTHROTTLE_MAX_DELAY = 60
# The average number of requests Scrapy should be sending in parallel to
# each remote server
# AUTOTHROTTLE_TARGET_CONCURRENCY = 1.0
# Enable showing throttling stats for every response received:
# AUTOTHROTTLE_DEBUG = False

# Enable and configure HTTP caching (disabled by default)
# See https://docs.scrapy.org/en/latest/topics/downloader-middleware.html#httpcache-middleware-settings
# HTTPCACHE_ENABLED = True
# HTTPCACHE_EXPIRATION_SECS = 0
# HTTPCACHE_DIR = 'httpcache'
# HTTPCACHE_IGNORE_HTTP_CODES = []
# HTTPCACHE_STORAGE = 'scrapy.extensions.httpcache.FilesystemCacheStorage'
EXTRACTOR_ADDR = "http://extractor-link-server.thanos.sogou/api/"



import base64
import logging
import os

import ujson as json
from aiohttp import ClientSession
from scrapy.utils.python import to_unicode
from scrapy.utils.url import parse_url

from lich_linkextractor.utils import as_deferred

logger = logging.getLogger(__name__)

DEPTH_LIMIT_KEY = "extractor.depth_limit"
DROP_ANCHOR_KEY = "extractor.drop_anchor"
key_map = {
    "allow": "AllowedURLFilters",
    "deny": "DisallowedURLFilters",
    "allow_domains": "AllowedDomains",
    "deny_domains": "DisallowedDomains",
    "restrict_xpaths": "XPathQuerys",
    "restrict_css": "CSSSelectors",
    "same_domain_only": "OnlyHomeSite",
}


class AyncExtractor:
    def __init__(self, depth_limit=None, server_addr=None):
        self.depth_limit = depth_limit
        self.server_addr = server_addr
        self.client = ClientSession()

    @classmethod
    def from_crawler(cls, crawler):
        crawler.settings
        depth_limit = crawler.settings.get("DEPTH_LIMIT")
        depth_limit = int(depth_limit) if depth_limit else None
        server_addr = crawler.settings.get("EXTRACTOR_ADDR")
        return cls(depth_limit=depth_limit, server_addr=server_addr)

    def process_response(self, response):
        d = as_deferred(self.extract_links_and_extracts(response))

        def _on_success(extract_recode):
            links = extract_recode["links"]
            extracts = extract_recode["extracts"]
            if links is not None and len(links) > 0:
                assert "extractor.links" not in response.meta
                response.meta["extractor.links"] = links
            if extracts is not None and len(extracts) > 0:
                assert "extractor2.extracts" not in response.meta
                response.meta["extractor2.extracts"] = extracts

        d.addCallback(_on_success)
        d.addBoth(lambda _: response)
        return d

    def _get_base_ex_req(self, response):
        base_ex_req = None
        base_ex_req = {
            "URL": response.url,
            "Content": to_unicode(base64.encodebytes(response.body)),
            # "Content": response.body.decode("utf8"),
            "ContentType": "text/html; charset=utf-8",
        }
        return base_ex_req

    async def _get_links(self, rules, response):
        res = []
        base_ex_req = self._get_base_ex_req(response)
        if base_ex_req is None:
            return res
        #default values
        base_ex_req[key_map["same_domain_only"]]= False
        base_ex_req[key_map["allow_domains"]]=[]

        for rule in rules:
            ex_req = base_ex_req
            for key in [
                "allow",
                "deny",
                "allow_domains",
                "deny_domains",
                "restrict_xpaths",
                "restrict_css",
                "same_domain_only",
            ]:
                v = rule.get(key, None)
                if v is not None and (
                    isinstance(v, str)
                    or (
                        isinstance(v, (list, tuple))
                        and all(isinstance(i, str) for i in v)
                    )
                    or isinstance(v, bool)
                ):
                    ex_req[key_map[key]] = v

            redirect_urls = response.request.meta.get("redirect_urls", None)
            extractor_redirect_switch = response.request.meta.get(
                "extractor.redirect.switch", False
            )
            if (
                ex_req[key_map["same_domain_only"]]
                and redirect_urls is not None
                and len(redirect_urls) > 0
                and not extractor_redirect_switch
            ):
                root_host = parse_url(redirect_urls[0]).netloc.lower()
                if root_host and isinstance(ex_req[key_map["allow_domains"]], list):
                    ex_req[key_map["allow_domains"]].append(root_host)
                    ex_req[key_map["same_domain_only"]] = False
            resp = await self._send(self.server_addr, json.dumps(ex_req), 3)
            s = set()
            for value in resp.values():
                for link in value:
                    s.add(link)
            res.append(list(s))
        return res

    async def _get_extracts(self, extract_template, response):
        base_ex_req = self._get_base_ex_req(response)
        if base_ex_req is None:
            return res
        ex_req = base_ex_req
        ex_req["template"] = extract_template
        resp = await self._send(
            os.path.join(self.server_addr, "extract"), json.dumps(ex_req), 3
        )
        return resp["extract"]

    def _get_max_depth(self, req_depth_limit):
        if self.depth_limit is None:
            return req_depth_limit
        if req_depth_limit is None:
            return self.depth_limit
        return min(self.depth_limit, req_depth_limit)

    async def extract_links_and_extracts(self, response):
        extract_recode = {}
        meta = response.request.meta
        extracts = None
        links = None
        extract_template = response.request.meta.get("extractor2.template", None)
        if extract_template is not None and isinstance(extract_template, dict):
            extracts = await self._get_extracts(extract_template, response)
        rules = response.request.meta.get("extractor.rules", None)
        if rules is not None and isinstance(rules, list):
            # check depth limit
            depth = int(meta["depth"]) if "depth" in meta else 0
            req_depth_limit = (
                int(meta[DEPTH_LIMIT_KEY]) if DEPTH_LIMIT_KEY in meta else None
            )
            max_depth = self._get_max_depth(req_depth_limit)
            if not (max_depth!=None and depth!=None and depth >= max_depth):
                links = await self._get_links(rules, response)
        extract_recode["links"] = links
        extract_recode["extracts"] = extracts
        return extract_recode

    async def _send(self, addr, data, num_retries):
        if addr != None and data != None:
            async with self.client.post(addr, data=data) as response:
                if response.status == 200:
                    return await response.json()
                if response.status == 404:
                    if num_retries > 0:
                        return await self._send(addr, data, num_retries - 1)
                response.raise_for_status()

    async def spider_closed(self):
        return await self.client.close()



import os
import subprocess
import sys


def test_scrapy_check():
    run_thru_shell = sys.platform.startswith("win")
    proc = subprocess.Popen(["scrapy", "check"], shell=run_thru_shell, cwd=os.getcwd(),)
    exit_status = proc.wait()
    assert exit_status == 0



# -*- coding: utf-8 -*-

from scrapy import signals

from lich_linkextractor.async_extension import AyncExtractor


class LinkExtractorDownloaderMiddleware:
    def __init__(self, crawler):
        self.async_extension = AyncExtractor.from_crawler(crawler)
        crawler.signals.connect(self.spider_closed, signals.spider_closed)

    def process_response(self, request, response, spider):
        response.request = request
        return self.async_extension.process_response(response)

    def spider_closed(self, spider):
        return self.async_extension.spider_closed()

    @classmethod
    def from_crawler(cls, crawler):
        middleware = cls(crawler)
        return middleware



from setuptools import find_packages, setup


def read(*filenames, **kwargs):
    import io
    from os.path import dirname, join

    encoding = kwargs.get("encoding", "utf-8")
    sep = kwargs.get("sep", "\n")
    buf = []
    for filename in filenames:
        with io.open(join(dirname(__file__), filename), encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


setup(
    name="lich_linkextractor",
    version="0.0.10",
    packages=find_packages(),
    include_package_data=True,
    license="MIT License",
    description="Auto Generated by os-scrapy-cookiecutter",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=open("requirements/requirements.txt").read().split("\n"),
    python_requires=">=3.6",
    zip_safe=False,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



# -*- coding: utf-8 -*-

# Define here the models for your scraped items
#
# See documentation in:
# https://docs.scrapy.org/en/latest/topics/items.html

import scrapy


class LichLinkextractorItem(scrapy.Item):
    # define the fields for your item here like:
    # name = scrapy.Field()
    pass


class ExampleItem(scrapy.Item):
    # auto generated by os-scrapy-cookiecutter
    url = scrapy.Field()
    request_headers = scrapy.Field()
    response_headers = scrapy.Field()
    status = scrapy.Field()
    meta = scrapy.Field()
    body = scrapy.Field()
